/**
 *  
 */
var request = require('request');
var requestp = require('request-promise');
var path = require('path');
var Promise = require('bluebird');
var fs = Promise.promisifyAll(require("fs"));
var deferred = require('deferred');
var postAction = require('./postAction');
var filter = require("./filter");   // filter processing
var galaxy = require("./galaxyUtils");
var util = require("./utils");
//var prettyjson = require('prettyjson');   // for debugging

module.exports = {

    initialize: function(cb) {
        // todo: check that galaxy is running

        galaxy.init(function(history) {

            historyId = history.historyId;

        }, function(err) {
            sails.log.error("failed galaxy.init",err);
        });


        sails.on('hook:orm:loaded', function() {
            // do something after hooks are loaded
            return cb();
        });
    },

    workflowSubmit: function (req, res, next) {
        sails.log.info("JBlast POST /jbapi/workflowsubmit",req.body);
        var params = {
          region:req.body.region,
          workflow:req.body.workflow,
          dataSetPath:req.body.dataSetPath,
          monitorFn:monitorWorkflow
        };
        galaxy.workflowSubmit(params,
          function(data,err) {
              if (err !== null) {
                  sails.hooks['jbcore'].resSend(res,err);
              }
              sails.hooks['jbcore'].resSend(res,data);
          });
    },

    getWorkflows: function (req, res, next) {
          sails.log("JBlast GET /jbapi/getworkflows");

          galaxy.galaxyGET("/api/workflows",function(workflows,err) {
              if (err !== null) {
                  return res.send({status:'error',msg:"galaxy GET /api/workflows failed",err:err});
              }
              return res.send(workflows);
          });
    },
    /** post /jbapi/setfilter - send filter parameters
     * 
     * @param {type} req
     *    data = req.body
     *    data.filterParams = {score:{val: 50}, evalue:{val:-2}...
     *    data.dataSet = (i.e. "sample_data/json/volvox" generally from config.dataRoot)
     *    data.asset = 
     * @param {type} res
     * @param {type} next
     * @returns {undefined}
     */

    setFilter: function (req, res, next) {
        sails.log.info("JBlast","POST /jbapi/setfilter",req.body);
        rest_applyFilter(req,res);
    },
    /**
     * Get info about the given track
     */
    getBlastData: function (req, res, next) {
        sails.log.info("JBlast","/jbapi/getblastdata");
        rest_applyFilter(req,res);
    },
    getTrackData: function (req, res, next) {
          sails.log("JBlast /jbapi/gettrackdata called");
          var params = req.allParams();
          sails.log('asset',req.param('asset'));
          sails.log('dataset',req.param('dataset'));
          //sails.log('req.allParams()',req.allParams());

          var asset = req.param('asset');
          var dataset = req.param('dataset');

          var g = sails.config.globals.jbrowse;

          //var gfffile = g.jbrowsePath + dataset +'/'+ g.jblast.blastResultPath + '/' + 'sampleResult.gff3';
          var gfffile = g.jbrowsePath + dataset + '/'+ g.jblast.blastResultPath + '/' + asset +'.gff3';

          try {
              var content = fs.readFileSync(gfffile);
          }
          catch (err) {
              var str = JSON.stringify(err);
              //var str = str.split("\n");
              sails.log.error("failed to retrieve gff3 file",str);
              return sails.hooks['jbcore'].resSend(res,{status: 'error', msg: str, err:err});
          };

          return res.send(content);
    },
    /**
     * Return hits data given hit key
     */
    getHitDetails: function (req, res, next) {
          sails.log("JBlast /jbapi/gethitdetails called");
          //todo: handle errors
          rest_getHitDetails(req,res,function(hitData,err) {
              return res.send(hitData);
          });
    },
    /**
     * returns accession data given accesion number.
     * Utilizes Entrez service
     */
    lookupAccession: function (req, res, next) {
        sails.log("JBlast /jbapi/lookupaccession called");

        function accessionLookup(req,res) {
            this.accession.lookup(req,res,function(data,err) {
                res.send(data);
            });
            
        }
        // load accession module only on first time call
        if (typeof this.accession === 'undefined') {
            
            var g = sails.config.globals.jbrowse;
            var accModule = g.accessionModule;

            if (typeof accModule === 'undefined') accModule = "./accessionEntrez";

            this.accession = require(accModule);
            
            this.accession.init(req,res,function() {
                accessionLookup(req,res);
            });
        }
        else {
            accessionLookup(req,res);
        }
    }

};

/**
 * Process REST /jbapi/gethitdetails
 * @param {type} req
 * @param {type} res
 * @param {type} cb
 * @returns {undefined}
 */

function rest_getHitDetails(req,res, cb) {
    var asset = req.param('asset');
    var hitkey = req.param('hitkey');
    var dataset = req.param('dataset');
    filter.getHitDetails(hitkey, dataset, asset, function(hitData) {
       cb(hitData); 
    });
};

function rest_applyFilter(req,res) {
    sails.log.debug("rest_applyFilter()");
    var g = sails.config.globals;
    var requestData = req.body;

    // get /jbapi/getdata is used then the asset and dataset will be in th params, so extract them.
    var asset = req.param('asset');
    var dataSet = req.param('dataset');

    if (typeof asset !== 'undefined') requestData.asset = asset;
    if (typeof dataSet !== 'undefined') requestData.dataSet = dataSet;
    
    sails.log.debug("requestData",requestData);

    var err = filter.writeFilterSettings(requestData,function(filterData) {
        filter.applyFilter(filterData,requestData,function(data) {
    
            return res.send(data);
        });
    });
    if (err) {
        return res.send({status:'error',err:err});
    }
};
    
/**
 * Monitor workflow
 * @param {type} kWorkflowJob
 * @returns {undefined}
 */

function monitorWorkflow(kWorkflowJob){
    var wId = kWorkflowJob.data.workflow.workflow_id;
    sails.log.debug('monitorWorkflow starting, wId',wId);
    
    var timerloop = setInterval(function(){
        var hId = kWorkflowJob.data.workflow.history_id;
        
        //todo: if workflow fails, output will not exist.  Need to handle this.
        var outputs = kWorkflowJob.data.workflow.outputs;    // list of workflow output history ids
        var outputCount = outputs.length;
        
        sails.log.info ("history",hId);
        
        // get history entries
        var url = '/api/histories/'+hId+'/contents';
        galaxy.galaxyGET(url,function(hist,err) {

            if (err !== null) {
                var msg = wId + " monitorWorkflow: failed to get history "+hId;
                sails.log.error(msg,err);
                clearInterval(timerloop);
                kWorkflowJob.kDoneFn(new Error(msg));
                return;
            }
            // reorg to assoc array
            var hista = {};
            for(var i in hist) hista[hist[i].id] = hist[i];

            // determine aggregate state
            var okCount = 0;
            for(var i in outputs) {
                // if any are running or uploading, we are active
                if(hista[outputs[i]].state==='running' || hista[outputs[i]].state==='upload')
                    break;
                // if something any history error, the whole workflow is in error
                if(hista[outputs[i]].state==='error') {
                    clearInterval(timerloop);
                    kWorkflowJob.state('failed');
                    kWorkflowJob.save();
                    sails.log.debug(wId,'workflow completed in error');
                    break;
                }
                if(hista[outputs[i]].state==='ok')
                    okCount++;
            }
            sails.log.debug(wId,'workflow step',okCount,'of',outputCount);
            
            kWorkflowJob.progress(okCount,outputCount+1,{workflow_id:wId});

            // complete if all states ok
            if (outputCount === okCount) {
                clearInterval(timerloop);
                kWorkflowJob.state('complete');
                kWorkflowJob.save();
                sails.log.debug(wId,'workflow completed');
                setTimeout(function() {
                    postAction.doCompleteAction(kWorkflowJob,hista);            // workflow completed
                },10);
            }
        });
        
    },3000);
};
